from sqlite3 import Connection
from pathlib import Path


SOUND_BUILD_SCRIPT = '''
-- Sound
CREATE TABLE SoundFile (
    SoundFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ResourcePack_fk INTEGER,

    path Path NOT NULL,
    -- The identifier is the path without extension. This is added to the DB to
    -- make searches easier.
    identifier TEXT NOT NULL,
    FOREIGN KEY (ResourcePack_fk) REFERENCES ResourcePack (ResourcePack_pk)
        ON DELETE CASCADE
);
CREATE INDEX SoundFile_ResourcePack_fk
ON SoundFile (ResourcePack_fk);
'''

def load_sounds(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM ResourcePack WHERE ResourcePack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    for sound_path in (rp_path / "sounds").rglob("*.wav"):
        load_sound(db, sound_path, rp_path, rp_id)
    for sound_path in (rp_path / "sounds").rglob("*.ogg"):
        load_sound(db, sound_path, rp_path, rp_id)
    for sound_path in (rp_path / "sounds").rglob("*.fsb"):
        load_sound(db, sound_path, rp_path, rp_id)

def load_sound(db: Connection, sound_path: Path, rp_path: Path, rp_id: int):
    cursor = db.cursor()
    # SOUND FILE AND ITS IDENTIFIER
    cursor.execute(
        """
        INSERT INTO SoundFile (
            path, identifier, ResourcePack_fk
        ) VALUES (?, ?, ?)
        """,
        (
            sound_path.as_posix(),
            sound_path.relative_to(rp_path).with_suffix("").as_posix(),
            rp_id
        )
    )


from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
import json

ATTACHABLE_BUILD_SCRIPT = '''
-- Attachable
CREATE TABLE AttachableFile (
    AttachableFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ResourcePack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (ResourcePack_fk) REFERENCES ResourcePack (ResourcePack_pk)
        ON DELETE CASCADE
);
CREATE INDEX AttachableFile_ResourcePack_fk
ON AttachableFile (ResourcePack_fk);

CREATE TABLE Attachable (
    Attachable_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    AttachableFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,

    FOREIGN KEY (AttachableFile_fk) REFERENCES AttachableFile (AttachableFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX Attachable_AttachableFile_fk
ON Attachable (AttachableFile_fk);

-- This maps the item used by the attachable. Attachables have two types of
-- connecting items. They can either have an identifier matching to the item
-- or they can have the 'item' property which maps the item with a condition.
-- In both cases this field is added to the database.
CREATE TABLE AttachableItemField (
    AttachableItemField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    Attachable_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    condition TEXT,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (Attachable_fk) REFERENCES Attachable (Attachable_pk)
        ON DELETE CASCADE
);
CREATE INDEX AttachableItemField_Attachable_fk
ON AttachableItemField (Attachable_fk);

CREATE TABLE AttachableMaterialField (
    AttachableMaterialField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    Attachable_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (Attachable_fk) REFERENCES Attachable (Attachable_pk)
        ON DELETE CASCADE
);
CREATE INDEX AttachableMaterialField_Attachable_fk
ON AttachableMaterialField (Attachable_fk);

CREATE TABLE AttachableTextureField (
    AttachableTextureField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    Attachable_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (Attachable_fk) REFERENCES Attachable (Attachable_pk)
        ON DELETE CASCADE
);
CREATE INDEX AttachableTextureField_Attachable_fk
ON AttachableTextureField (Attachable_fk);

CREATE TABLE AttachableGeometryField (
    AttachableGeometryField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    Attachable_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (Attachable_fk) REFERENCES Attachable (Attachable_pk)
        ON DELETE CASCADE
);
CREATE INDEX AttachableGeometryField_Attachable_fk
ON AttachableGeometryField (Attachable_fk);


CREATE TABLE AttachableRenderControllerField (
    AttachableRenderControllerField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    Attachable_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    condition TEXT,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (Attachable_fk) REFERENCES Attachable (Attachable_pk)
        ON DELETE CASCADE
);
CREATE INDEX AttachableRenderControllerField_Attachable_fk
ON AttachableRenderControllerField (Attachable_fk);


CREATE TABLE AttachableAnimationField (
    AttachableAnimationField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    Attachable_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (Attachable_fk) REFERENCES Attachable (Attachable_pk)
        ON DELETE CASCADE
);
CREATE INDEX AttachableAnimationField_Attachable_fk
ON AttachableAnimationField (Attachable_fk);

CREATE TABLE AttachableAnimationControllerField (
    AttachableAnimationControllerField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    Attachable_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (Attachable_fk) REFERENCES Attachable (Attachable_pk)
        ON DELETE CASCADE
);
CREATE INDEX AttachableAnimationControllerField_Attachable_fk
ON AttachableAnimationControllerField (Attachable_fk);
'''

def load_attachables(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM ResourcePack WHERE ResourcePack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    for attachable_path in (rp_path / "attachables").rglob("*.json"):
        load_attachable(db, attachable_path, rp_id)

def load_attachable(db: Connection, attachable_path: Path, rp_id: int):
    cursor = db.cursor()
    # ATTACHABLE FILE
    cursor.execute(
        "INSERT INTO AttachableFile (path, ResourcePack_fk) VALUES (?, ?)",
        (attachable_path.as_posix(), rp_id)
    )
    file_pk = cursor.lastrowid
    try:
        walker = load_jsonc(attachable_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return
    # ATTACHABLE
    description = walker / "minecraft:attachable" / "description"
    identifier_walker = description / 'identifier'
    identifier_data = identifier_walker.data
    if not isinstance(identifier_data, str):
        return
    cursor.execute(
        '''
        INSERT INTO Attachable (
            identifier, AttachableFile_fk
        ) VALUES (?, ?)
        ''',
        (identifier_data, file_pk)
    )
    attachable_pk = cursor.lastrowid

    # ATTACHABLE ITEM FIELD
    items = description / "item" // str
    if len(items) == 0:
        cursor.execute(
            '''
            INSERT INTO AttachableItemField (
                identifier, Attachable_fk, jsonPath
            ) VALUES (?, ?, ?)
            ''',
            (identifier_data, attachable_pk, identifier_walker.path_str)
        )
    else:
        for item in items:
            if not isinstance(item.data, str):
                continue
            cursor.execute(
                '''
                INSERT INTO AttachableItemField (
                    identifier, Attachable_fk, condition,
                    jsonPath
                ) VALUES (?, ?, ?, ?)
                ''',
                (
                    item.parent_key, attachable_pk,
                    item.data, item.path_str
                )
            )

    # ATTACHABLE MATERIAL FIELD
    materials = description / "materials" // str
    for material in materials:
        if not isinstance(material.data, str):
            continue
        cursor.execute(
            '''
            INSERT INTO AttachableMaterialField (
                shortName, identifier, Attachable_fk, jsonPath
            ) VALUES (?, ?, ?, ?)
            ''',
            (
                material.parent_key, material.data, attachable_pk,
                material.path_str
            )
        )

    # ATTACHABLE TEXTURE FIELD
    textures = description / "textures" // str
    for texture in textures:
        if not isinstance(texture.data, str):
            continue
        cursor.execute(
            '''
            INSERT INTO AttachableTextureField (
                shortName, identifier, Attachable_fk, jsonPath
            ) VALUES (?, ?, ?, ?)
            ''',
            (
                texture.parent_key, texture.data, attachable_pk,
                texture.path_str
            )
        )

    # ATTACHABLE GEOMETRY FIELD
    geometries = description / "geometry" // str
    for geometry in geometries:
        if not isinstance(geometry.data, str):
            continue
        cursor.execute(
            '''
            INSERT INTO AttachableGeometryField (
                shortName, identifier, Attachable_fk, jsonPath
            ) VALUES (?, ?, ?, ?)
            ''',
            (
                geometry.parent_key, geometry.data, attachable_pk,
                geometry.path_str
            )
        )

    # ATTACHABLE RENDER CONTROLLER FIELD
    render_controllers = description / "render_controllers" // int
    for render_controller in render_controllers:
        if isinstance(render_controller.data, str):
            cursor.execute(
                '''
                INSERT INTO AttachableRenderControllerField (
                    identifier, Attachable_fk, jsonPath
                ) VALUES (?, ?, ?)
                ''',
                (
                    render_controller.data, attachable_pk,
                    render_controller.path_str
                )
            )
        else:
            # Render contoroller can be an object with pair of values, name
            # and condition.
            for render_controller in render_controller // str:
                cursor.execute(
                    '''
                    INSERT INTO AttachableRenderControllerField (
                        identifier, condition, Attachable_fk, jsonPath
                    ) VALUES (?, ?, ?, ?)
                    ''',
                    (
                        render_controller.parent_key, render_controller.data,
                        attachable_pk, render_controller.path_str
                    )
                )
    # ANIMATIONS & ANIMATION CONTROLLERS
    for animation in description / "animations" // str:
        if isinstance(animation.data, str):
            identifier = animation.data
        else:
            continue
        if animation.data.startswith("controller.animation."):
            # Animations
            cursor.execute(
                '''
                INSERT INTO AttachableAnimationControllerField (
                    Attachable_fk, shortName, identifier, jsonPath
                ) VALUES (?, ?, ?, ?)
                ''',
                (attachable_pk, animation.parent_key, identifier, animation.path_str))
        elif animation.data.startswith("animation."):
            # Animation Controllers

            cursor.execute(
                '''
                INSERT INTO AttachableAnimationField (
                    Attachable_fk, shortName, identifier, jsonPath
                ) VALUES (?, ?, ?, ?)
                ''',
                (attachable_pk, animation.parent_key, identifier, animation.path_str))



from __future__ import annotations
import json
import re
from typing import Union, Type, Optional, IO, Callable, Iterator

class SKIP_LIST:
    '''Used as literal value for JSONSplitWalker paths'''

# def _remove_escape_characters(text: str) -> str:
#     '''Prints to a string, removint the escape characters'''
#     with io.StringIO() as output:
#         print(text, file=output, end='')
#         contents = output.getvalue()
#     return contents

def _tuple_to_path_str(path: tuple[Union[str, int], ...]):
    result = []
    for k in path:
        if isinstance(k, int):
            result.append(f'[{k}]')
        elif isinstance(k, str):
            if re.fullmatch("[a-zA-Z$_]+[a-zA-Z$_0-9]*", k):
                # Mathes JS variable name (like connect like a.b.c)
                if len(result) == 0:  # First item skip the dot
                    result.append(k)
                else: # Add dot before
                    result.append(f".{k}")
            else:
                # Does not match JS variable name (like connect like
                # ["a"]["b"]["c"])
                k = json.dumps(k)  # escape special characters and add quotes
                result.append(f'[{k}]')
        else:
            raise TypeError(f"Invalid key type {type(k)}")
    return "".join(result)

class JSONPath:
    '''
    Represents a path in a JSON file. The paths internally use a tuple listing
    the keys, but they can be represented or created from a string. The string
    representation is similar to the one used in JavaScript.

    The path objects can be used to access the data of :class:`JSONWalker`.

    Example:
        >>> from better_json_tools import JSONPath
        >>> path = JSONPath(("a", "$abc", 1, 2, 'with quote "')) 
        >>> print(path.data)
        ... ('a', '$abc', 1, 2, 'with quote "')
        >>> print(path)
        ... a.$abc[1][2]["with quote \""]
        >>> another_path = JSONPath(str(path))
        >>> print(another_path)
        ... a.$abc[1][2]["with quote \""]
        >>> print(another_path.data == path.data)
        ... True
    '''
    def __init__(self, path: Union[str, tuple[Union[str, int], ...]]):
        if isinstance(path, str):
            self.data = JSONPath._from_path_str(path)
        else:
            self.data = path

    def __str__(self) -> str:
        '''Returns a string representation of the path.'''
        return _tuple_to_path_str(self.data)

    @staticmethod
    def _from_path_str(path_str: str) -> tuple[Union[str, int], ...]:
        '''Converts a path string to a JSONPath.'''
        if path_str == "":
            return tuple()

        # Results
        path = []
        curr_path = path_str

        # Add . to the start of the string to make the patterm matching work
        # better
        if not curr_path.startswith("["):
            curr_path = "." + curr_path
        while True:
            if curr_path == "":
                break
            if curr_path.startswith("."):
                # Match a.b.c
                match = re.match(r"\.([a-zA-Z$_]+[a-zA-Z$_0-9]*)", curr_path)
                if match is None:
                    raise ValueError(f"Invalid path: {path_str}")
                path.append(match.group(1))
                curr_path = curr_path[match.end():]
            elif curr_path.startswith("["):
                if len(curr_path) < 3: # shortest possible path is like [0]
                    raise ValueError(f"Invalid path: {path_str}")
                if curr_path[1] == '"':
                    # Match ["a"]["b"]["c"]
                    match = re.match(r'\[("(?:[^"]|\\")*")\]', curr_path)
                    if match is None:
                        raise ValueError(f"Invalid path: {path_str}")
                    path.append(json.loads(match.group(1)))
                    curr_path = curr_path[match.end():]
                else:
                    # Match [0][1][2]
                    match = re.match(r"\[([0-9]+)\]", curr_path)
                    if match is None:
                        raise ValueError(f"Invalid path: {path_str}")
                    path.append(int(match.group(1)))
                    curr_path = curr_path[match.end():]
            else:
                raise ValueError(f"Invalid path: {path_str}")
        return tuple(path)

## Type definitions
JSON = Union[dict, list, str, float, int, bool, None]
JSON_KEY = Union[str, int]
JSON_PATH_KEY = Union[str, int, JSONPath]
JSON_SPLIT_KEY = Union[str, Type[int], Type[str], None, Type[SKIP_LIST]]
JSON_WALKER_DATA = Union[dict, list, str, float, int, bool, None, Exception]



class JSONWalker:
    '''
    A class that represents a path in the JSON file for easy access to its
    values.
    '''
    def __init__(
            self, data: JSON_WALKER_DATA, *,
            parent: Optional[JSONWalker] = None,
            parent_key: Optional[JSON_KEY] = None):
        if not isinstance(
                data, (Exception, dict, list, str, float, int, bool, type(None))):
            raise ValueError('Input data is not JSON.')
        self._data: JSON_WALKER_DATA = data
        self._parent = parent
        self._parent_key = parent_key

    @property
    def parent(self) -> JSONWalker:
        '''
        The parent of this json walker (the walker that created this walker).

        :rises: :class:`KeyError` when this :class:`JSONWalker` is a root
            object.
        '''
        if self._parent is None:
            raise KeyError("You can't get parent of the root object.")
        return self._parent

    @property
    def parent_key(self) -> JSON_KEY:
        '''
        The key used to access this walker from its parent

        :rises: :class:`KeyError` when this :class:`JSONWalker` is a root
            object
        '''
        if self._parent_key is None:
            raise KeyError("You can't get parent of the root object.")
        return self._parent_key

    @staticmethod
    def loads(json_text: Union[str, bytes], **kwargs) -> JSONWalker:
        '''
        Creates json walker using `json.loads()` function. Passes all arguments
        to `json.loads` and tries to creat the walker base on the result.
        '''
        data = json.loads(json_text, **kwargs)
        return JSONWalker(data)

    @staticmethod
    def load(json_file: IO, **kwargs) -> JSONWalker:
        '''
        Creates json walker using `json.load()` function. Passes all arguments
        to `json.load` and tries to creat the walker base on the result.
        '''
        data = json.load(json_file, **kwargs)
        return JSONWalker(data)

    @property
    def data(self) -> JSON_WALKER_DATA:
        '''
        The data from JSON that this walker points to.
        '''
        return self._data

    @data.setter
    def data(self, value: JSON):
        if self._parent is not None:
            self.parent.data[  # type: ignore
                self.parent_key  # type: ignore
            ] = value
        self._data = value

    def create_path(
            self, data: JSON, *,
            exists_ok: bool = True,
            can_break_data_structure: bool = True,
            can_create_empty_list_items: bool = True,
            empty_list_item_factory: Optional[Callable[[], JSON]] = None):
        '''
        Creates path to the part of JSON file pointed by this walker.

        :param data: the data to put at the end of the path.
        :param exists_ok: if False, the ValueError will be risen if the path
            to this item already exists.
        :param can_break_data_structure: if True than the function will be able
            to replace certain existing paths with dicts or lists. Example -
            if path "a"/"b"/"c" points at integer, creating path
            "a"/"b"/"c"/"d" will replace this integer with a dict in order to
            make "d" a valid key. Setting this to false would cause a KeyError
            in this situation.
        :param can_create_empty_list_items: enables filling up the lists in
            JSON with values produced by the empty_list_item_factory in order
            to match the item index specified in the path. Example - if you
            specify a path to create "a"/5/"c" but the list at "a" path only
            has 2 items, then the function will create additional item so
            the 5th index can be valid.
        :param empty_list_item_factory: a function used to create items for
            lists in order to make indices specified in the path valid (see
            can_create_empty_list_items function parameter). If this value
            is left as None than the lists will be filled with null values.
        '''
        if self.exists:
            if exists_ok:
                return
            raise ValueError("Path already exists")
        if empty_list_item_factory is None:
            empty_list_item_factory = lambda: None
        curr_item = self.root
        path = self.path
        for key in path:
            if isinstance(key, str):  # key is a string data must be a dict
                if not isinstance(curr_item.data, dict):
                    if not can_break_data_structure:
                        raise KeyError(key)
                    curr_item.data = {}
                if key not in curr_item.data:
                    can_break_data_structure = True  # Creating new data
                curr_item = curr_item / key
            elif isinstance(key, int):  # key is an int data must be a list
                if key < 0:
                    raise KeyError(key)
                if not isinstance(curr_item.data, list):
                    if not can_break_data_structure:
                        raise KeyError(key)
                    curr_item.data = []
                if len(curr_item.data)-1 < key:
                    if not can_create_empty_list_items:
                        raise KeyError(key)
                    curr_item.data += [
                        empty_list_item_factory()
                        for _ in range(1+key-len(curr_item.data))
                    ]
                    can_break_data_structure = True  # Creating new data
                curr_item = curr_item / key
            else:
                raise KeyError(key)
        self._parent = curr_item.parent
        self._parent_key = curr_item.parent_key
        self.data = data

    @property
    def exists(self) -> bool:
        '''
        Returns true if path to this item already exists. This function
        recursively checks the entire path to this item starting from root so
        even if the object is detached from the root somewhere in the middle
        of the path, the function will still return correct value.
        '''
        keys: list[JSON_KEY] = []
        root = self
        try:
            while True:
                keys.append(root.parent_key)
                root = root.parent
        except KeyError:
            pass
        keys = list(reversed(keys))
        root_data = root.data
        try:
            for key in keys:
                root_data = root_data[key]  # type: ignore
        except:  # pylint: disable=bare-except
            return False
        return True

    @property
    def root(self) -> JSONWalker:
        '''
        The root object of this JSON file.
        '''
        root = self
        try:
            while True:
                root = root.parent
        except KeyError:
            pass
        return root

    @property
    def path(self) -> tuple[JSON_KEY, ...]:
        '''
        Full JSON path up to this point starting from the root of the JSON file
        in from of a tuple of keys.
        '''
        result: list[JSON_KEY] = []
        parent = self
        try:
            while True:
                result.append(parent.parent_key)
                parent = parent.parent
        except KeyError:
            pass
        return tuple(reversed(result))

    @property
    def path_str(self) -> str:
        '''
        Full JSON path up to this point starting from the root of the JSON file
        in form of a string.
        '''
        return _tuple_to_path_str(self.path)

    def __truediv__(self, key: JSON_PATH_KEY) -> JSONWalker:
        '''
        The `/` operator creates descendant path in the JSON file.
        '''
        if isinstance(key, JSONPath):
            walker = self
            for k in key.data:
                walker = walker / k
            return walker
        else:
            try:
                return JSONWalker(
                    self.data[key],  # type: ignore
                    parent=self, parent_key=key)
            except Exception as e:  # pylint: disable=broad-except
                return JSONWalker(e, parent=self, parent_key=key)

    def __floordiv__(self, key: JSON_SPLIT_KEY) -> JSONSplitWalker:
        '''
        The `//` operator creates JSONSplitWalker object with multiple
        alternative paths that matched provided key.

        :raises:
            :class:`TypeError` - invalid input data type

            :class:`re.error` - invlid regular expression.
        '''
        # pylint: disable=too-many-return-statements
        # ANYTHING
        if key is None:
            if isinstance(self.data, dict):
                return JSONSplitWalker([
                    JSONWalker(v, parent=self, parent_key=k)
                    for k, v in self.data.items()
                ])
            if isinstance(self.data, list):
                return JSONSplitWalker([
                    JSONWalker(v, parent=self, parent_key=i)
                    for i, v in enumerate(self.data)
                ])
        # ANY LIST ITEM
        elif key is int:
            if isinstance(self.data, list):
                return JSONSplitWalker([
                    JSONWalker(v, parent=self, parent_key=i)
                    for i, v in enumerate(self.data)
                ])
        # ANY DICT ITEM
        elif key is str:
            if isinstance(self.data, dict):
                return JSONSplitWalker([
                    JSONWalker(v, parent=self, parent_key=k)
                    for k, v in self.data.items()
                ])
        # REGEX DICT ITEM
        elif isinstance(key, str):
            if isinstance(self.data, dict):
                result: list[JSONWalker] = []
                for k, v in self.data.items():
                    if re.fullmatch(key, k):
                        result.append(JSONWalker(
                            v, parent=self, parent_key=k))
                return JSONSplitWalker(result)
        # IF it's a list use ing key ELSE return split walker with self
        elif key is SKIP_LIST:
            if isinstance(self.data, list):
                return self // int
            return JSONSplitWalker([self])
        else:  # INVALID KEY TYPE
            raise TypeError(
                'Key must be a regular expression or one of the values: '
                'str, int, or None')
        # DATA DOESN'T ACCEPT THIS TYPE OF KEY
        return JSONSplitWalker([])

    def __add__(self, other: Union[JSONSplitWalker, JSONWalker]) -> JSONSplitWalker:
        '''
        The `+` operator adds json walkers creating a split walker with more
        values.
        '''
        if isinstance(other, JSONWalker):
            data = [self, other]
        else:
            data = other.data + [self]
        return JSONSplitWalker(
            [i for i in data if not isinstance(i.data, Exception)])


class JSONSplitWalker:
    '''
    Multiple walker objects grouped together. This class can be browse JSON
    file contents from multiple JSON paths at once.
    '''

    def __init__(self, data: list[JSONWalker]) -> None:
        self._data: list[JSONWalker] = data

    @property
    def data(self) -> list[JSONWalker]:
        '''
        The list of the :class:`JSONWalker` objects contained in this object.
        '''
        return self._data

    def __truediv__(self, key: JSON_PATH_KEY) -> JSONSplitWalker:
        '''
        Applies `/` operator to all of the :class:`JSONWalkers` in this split
        walker.
        '''
        result = []
        for walker in self.data:
            new_walker = walker / key
            if not isinstance(new_walker.data, Exception):
                result.append(new_walker)
        return JSONSplitWalker(result)

    def __floordiv__(self, key: JSON_SPLIT_KEY) -> JSONSplitWalker:
        '''
        Applies `//` operator to all of the :class:`JSONWalkers` in this split
        walker, creating even more split walkers (all groupped together in
        one object).
        '''
        result: list[JSONWalker] = []
        for walker in self.data:
            new_walker = walker // key
            result.extend(new_walker.data)
        return JSONSplitWalker(result)

    def __add__(self, other: Union[JSONSplitWalker, JSONWalker]) -> JSONSplitWalker:
        '''
        The `+` operator adds json walkers creating a split walker with more
        values.
        '''
        if isinstance(other, JSONWalker):
            data = self.data + [other]
        else:
            data = self.data + other.data
        return JSONSplitWalker(
            [i for i in data if not isinstance(i.data, Exception)])

    def __iter__(self) -> Iterator[JSONWalker]:
        '''
        Yield every walker contained in this object.
        '''
        for i in self.data:
            yield i

    def __len__(self) -> int:
        '''
        Return the number of walkers contained in this object.
        '''
        return len(self.data)



'''
Functions for analyzing Molang expressions.
'''
import re
from collections import defaultdict

def find_molang_resources(
        molang: str, resource_prefixes: list[str]) -> dict[str, list[str]]:
    '''
    Finds all resources of specified types in a molang expression. Returns
    a dictionary keyed by the resource type, with values being a list of
    found resources.

    Example
    -------
    >>> find_molang_resources(
            molang=(
                "q.is_baby = 1 ? array.skin[q.variant] : ("
                "q.mark_variant ? texture.default : texture.default2)"
            ),
            resource_prefixes=["array", "texture"])
    {'array': ['array.skin'], 'texture': ['texture.default', 'texture.default2']}
    '''
    molang = molang.lower()
    results: dict[str, list[str]] = {
        prefix: []
        for prefix in resource_prefixes
    }
    for resource_prefix in resource_prefixes:
        # Capturing groups crops the first part of the search for example:
        # geometry.default -> geometry
        item_pattern = re.compile(
            f'{resource_prefix}\\.(\\w+)', flags=re.IGNORECASE)
        for item_name in item_pattern.findall(molang):
            results[resource_prefix].append(item_name)
    return results



from pathlib import Path
from typing import Union
import json

from .compact_encoder import CompactEncoder
from .jsonc import JSONCDecoder
from .json_walker import JSONWalker, JSONSplitWalker, SKIP_LIST, JSONPath

VERSION = (1, 0, 0)  # COMPATIBILITY BREAK, NEW FEATURE, BUGFIX
__version__ = '.'.join([str(x) for x in VERSION])


def load_jsonc(jsonc_path: Union[Path, str]) -> JSONWalker:
    '''
    Loads JSONC file into a JSON walker object.
    '''
    if isinstance(jsonc_path, str):
        jsonc_path = Path(jsonc_path)
    try:
        with jsonc_path.open(encoding='utf8') as jsonc_file:
            data = json.load(jsonc_file)
    except json.JSONDecodeError as err:
        with jsonc_path.open(encoding='utf8') as jsonc_file:
            data = json.load(jsonc_file, cls=JSONCDecoder)
    return JSONWalker(data)





from typing import cast, Optional
from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
import json

SOUND_DEFINITIONS_BUILD_SCRIPT = '''
-- Sound definitions file
CREATE TABLE SoundDefinitionsFile (
    SoundDefinitionsFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ResourcePack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (ResourcePack_fk) REFERENCES ResourcePack (ResourcePack_pk)
        ON DELETE CASCADE
);
CREATE INDEX SoundDefinitionsFile_ResourcePack_fk
ON SoundDefinitionsFile (ResourcePack_fk);

CREATE TABLE SoundDefinition (
    SoundDefinition_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    SoundDefinitionsFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (SoundDefinitionsFile_fk) REFERENCES SoundDefinitionsFile (SoundDefinitionsFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX SoundDefinition_SoundDefinitionsFile_fk
ON SoundDefinition (SoundDefinitionsFile_fk);

CREATE TABLE SoundDefinitionSoundField (
    SoundDefinitionSoundField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    SoundDefinition_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (SoundDefinition_fk) REFERENCES SoundDefinition (SoundDefinition_pk)
        ON DELETE CASCADE
);
CREATE INDEX SoundDefinitionSoundField_SoundDefinition_fk
ON SoundDefinitionSoundField (SoundDefinition_fk);
'''


def load_sound_definitions(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM ResourcePack WHERE ResourcePack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    sound_definitions_path = rp_path / "sounds" / "sound_definitions.json"
    if sound_definitions_path.exists():
        load_sound_definition(db, sound_definitions_path, rp_id)


def load_sound_definition(db: Connection, sound_definition_path: Path, rp_id: int):
    cursor = db.cursor()
    # GEOMETRY FILE
    cursor.execute(
        "INSERT INTO SoundDefinitionsFile (path, ResourcePack_fk) VALUES (?, ?)",
        (sound_definition_path.as_posix(), rp_id)
    )
    file_pk = cursor.lastrowid
    try:
        sound_definitions_jsonc = load_jsonc(sound_definition_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return
    # Try to load from the "sound_definitions" key, if it doesn't exist,
    # load from the root of the file (old format)
    if (sound_definitions_jsonc / "sound_definitions").exists:
        sound_list = sound_definitions_jsonc / "sound_definitions" // str
        old_format = False
    else:
        sound_list = sound_definitions_jsonc // str
        old_format = True
    for sound_definition in sound_list:
        if old_format and sound_definition.parent_key == "format_version":
            continue
        sound_definition_identifier = cast(str, sound_definition.parent_key)
        cursor.execute(
            '''
            INSERT INTO SoundDefinition (
                identifier, jsonPath, SoundDefinitionsFile_fk
            ) VALUES (?, ?, ?)
            ''',
            (sound_definition_identifier, sound_definition.path_str, file_pk)
        )
        sound_definition_pk = cursor.lastrowid
        for sound in sound_definition / "sounds" // int:
            if isinstance(sound.data, str):
                sound_identifier = sound.data
            elif isinstance(sound.data, dict):
                sound_name_walker = sound / "name"
                if (
                        not sound_name_walker.exists or
                        not isinstance(sound_name_walker.data, str)):
                    continue
                sound_identifier = sound_name_walker.data
            else:
                continue
            cursor.execute(
                '''
                INSERT INTO SoundDefinitionSoundField (
                    identifier, jsonPath, SoundDefinition_fk
                ) VALUES (?, ?, ?)
                ''',
                (sound_identifier, sound.path_str, sound_definition_pk)
            )



from pathlib import Path
from sqlite3 import Connection
import sqlite3
from typing import Iterable, Literal

from .db_resource_pack import RESOURCE_PACK_BUILD_SCRIPT, load_resource_pack
from .db_geometry import GEOMETRY_BUILD_SCRIPT, load_geometries
from .db_client_entity import CLIENT_ENTITY_BUILD_SCRIPT, load_client_entities
from .db_render_controller import (
    RENDER_CONTROLLER_BUILD_SCRIPT, load_render_controllers)
from .db_texture import TEXTURE_BUILD_SCRIPT, load_textures
from .db_particle import PARTICLE_BUILD_SCRIPT, load_particles
from .db_rp_animation import RP_ANIMATION_BUILD_SCRIPT, load_rp_animations
from .db_rp_animation_controller import (
    RP_ANIMATION_CONTROLLER_BUILD_SCRIPT, load_rp_animation_controllers)
from .db_attachable import ATTACHABLE_BUILD_SCRIPT, load_attachables
from .db_sound_definitions import (
    SOUND_DEFINITIONS_BUILD_SCRIPT, load_sound_definitions)
from .db_sound import (
    SOUND_BUILD_SCRIPT, load_sounds)
from .db_behavior_pack import (
    BEHAVIOR_PACK_BUILD_SCRIPT, load_behavior_pack)


SCRIPT = '''
PRAGMA foreign_keys = ON;
'''

def _path_adapter(path: Path):
    return path.as_posix()


def _path_converter(path: bytes):
    return Path(path.decode('utf8'))

def create_db(db_path: str = ":memory:") -> Connection:
    '''
    Creates a new dtabase in :code:`db_path`. Runs all of the build scripts of
    the database components.

    :param db_path: The path to the database file. The argument is passed to
        :func:`sqlite3.connect` If the argument is :code:`":memory:"`, the
        database is created in memory. :code:`":memory:"` is the default value.
    '''
    sqlite3.register_adapter(Path, _path_adapter)
    sqlite3.register_converter("Path", _path_converter)
    db = sqlite3.connect(db_path, detect_types=sqlite3.PARSE_DECLTYPES)
    db.row_factory = sqlite3.Row

    db.executescript(SCRIPT)

    db.executescript(RESOURCE_PACK_BUILD_SCRIPT)
    db.executescript(CLIENT_ENTITY_BUILD_SCRIPT)
    db.executescript(RENDER_CONTROLLER_BUILD_SCRIPT)
    db.executescript(GEOMETRY_BUILD_SCRIPT)
    db.executescript(TEXTURE_BUILD_SCRIPT)
    db.executescript(PARTICLE_BUILD_SCRIPT)
    db.executescript(RP_ANIMATION_BUILD_SCRIPT)
    db.executescript(RP_ANIMATION_CONTROLLER_BUILD_SCRIPT)
    db.executescript(ATTACHABLE_BUILD_SCRIPT)
    db.executescript(SOUND_DEFINITIONS_BUILD_SCRIPT)
    db.executescript(SOUND_BUILD_SCRIPT)

    db.executescript(BEHAVIOR_PACK_BUILD_SCRIPT)
    return db

def open_db(db_path: str) -> Connection:
    '''
    Opens a database file. Usually these files are created by
    :func:`create_db`. This function doesn't check if the database has a valid
    structure. It assumes it does. This function only opens the database and
    sets some sqlite3 adapter and converter functions for Path objects.

    :param db_path: The path to the database file. The argument is passed to
        :func:`sqlite3.connect`.
    '''
    sqlite3.register_adapter(Path, _path_adapter)
    sqlite3.register_converter("Path", _path_converter)
    return sqlite3.connect(db_path, detect_types=sqlite3.PARSE_DECLTYPES)

DbRpItems = Literal[
    "geometries",
    "client_entities",
    "render_controllers",
    "textures",
    "particles",
    "rp_animations",
    "rp_animation_controllers",
    "attachables",
    "sound_definitions",
    "sounds",
]
'''
Possible values of :code:`include and :code:`exclude` arguments of
:func:`load_rp` function.
'''

def load_rp(
        db: Connection,
        rp_path: Path, *,
        include: Iterable[DbRpItems] = (
            "geometries",
            "client_entities",
            "render_controllers",
            "textures",
            "particles",
            "rp_animations",
            "rp_animation_controllers",
            "attachables",
            "sound_definitions",
            "sounds",
        ),
        exclude: Iterable[DbRpItems] = tuple()
    ) -> None:
    '''
    Loads resource pack data into the database.

    :param db: The database connection.
    :param rp_path: The path to the resource pack.
    :param include: A list of items to include. By default, all items are
        included.
    :param exclude: A list of items to exclude. By default, no items are
        excluded.

    If there is an item in both include and exclude, it is excluded. The
    include and exclude lists accept strings that are the names of the
    supported database components.
    '''
    rp_pk = load_resource_pack(db, rp_path)
    if (
            "geometries" in include and
            "geometries" not in exclude):
        load_geometries(db, rp_pk)
    if (
            "client_entities" in include and
            "client_entities" not in exclude):
        load_client_entities(db, rp_pk)
    if (
            "render_controllers" in include and
            "render_controllers" not in exclude):
        load_render_controllers(db, rp_pk)
    if (
            "textures" in include and
            "textures" not in exclude):
        load_textures(db, rp_pk)
    if (
            "particles" in include and
            "particles"  not in exclude):
        load_particles(db, rp_pk)
    if (
            "rp_animations" in include and
            "rp_animations"  not in exclude):
        load_rp_animations(db, rp_pk)
    if (
            "rp_animation_controllers" in include and
            "rp_animation_controllers"  not in exclude):
        load_rp_animation_controllers(db, rp_pk)
    if (
            "attachables" in include and
            "attachables"  not in exclude):
        load_attachables(db, rp_pk)
    if (
            "sound_definitions" in include and
            "sound_definitions"  not in exclude):
        load_sound_definitions(db, rp_pk)
    if (
            "sounds" in include and
            "sounds"  not in exclude):
        load_sounds(db, rp_pk)
    db.commit()

def load_bp(db: Connection, bp_path: Path) -> None:
    '''
    Loads behavior pack data into the database. Currently, this function only
    loads the Behavior Pack table and doesn't load any objects from the pack.

    :param db: The database connection.
    :param rp_path: The path to the resource pack.
    '''
    load_behavior_pack(db, bp_path)
    db.commit()



from typing import cast
from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
import json

RP_ANIMATION_CONTROLLER_BUILD_SCRIPT = '''
-- RpAnimationController
CREATE TABLE RpAnimationControllerFile (
    RpAnimationControllerFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ResourcePack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (ResourcePack_fk) REFERENCES ResourcePack (ResourcePack_pk)
        ON DELETE CASCADE
);
CREATE INDEX RpAnimationControllerFile_ResourcePack_fk
ON RpAnimationControllerFile (ResourcePack_fk);

CREATE TABLE RpAnimationController (
    RpAnimationController_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    RpAnimationControllerFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,
    
    FOREIGN KEY (RpAnimationControllerFile_fk) REFERENCES RpAnimationControllerFile (RpAnimationControllerFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX RpAnimationController_RpAnimationControllerFile_fk
ON RpAnimationController (RpAnimationControllerFile_fk);

CREATE TABLE RpAnimationControllerParticleEffect (
    RpAnimationControllerParticleEffect_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    RpAnimationController_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (RpAnimationController_fk) REFERENCES RpAnimationController (RpAnimationController_pk)
        ON DELETE CASCADE
);
CREATE INDEX RpAnimationControllerParticleEffect_RpAnimationController_fk
ON RpAnimationControllerParticleEffect (RpAnimationController_fk);

CREATE TABLE RpAnimationControllerSoundEffect (
    RpAnimationControllerSoundEffect_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    RpAnimationController_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (RpAnimationController_fk) REFERENCES RpAnimationController (RpAnimationController_pk)
        ON DELETE CASCADE
);
CREATE INDEX RpAnimationControllerSoundEffect_RpAnimationController_fk
ON RpAnimationControllerSoundEffect (RpAnimationController_fk);
'''

def load_rp_animation_controllers(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM ResourcePack WHERE ResourcePack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    for ac_path in (rp_path / "animation_controllers").rglob("*.json"):
        load_rp_animation_controller(db, ac_path, rp_id)


def load_rp_animation_controller(db: Connection, animation_controller_path: Path, rp_id: int):
    cursor = db.cursor()
    # RP ANIMATION FILE
    cursor.execute(
        "INSERT INTO RpAnimationControllerFile (path, ResourcePack_fk) VALUES (?, ?)",
        (animation_controller_path.as_posix(), rp_id)
    )
    file_pk = cursor.lastrowid
    try:
        acs_walker = load_jsonc(animation_controller_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return

    for ac_walker in acs_walker / "animation_controllers" // str:
        identifier_data: str = cast(str, ac_walker.parent_key)
        if not identifier_data.startswith("controller.animation."):
            continue
        cursor.execute(
            '''
            INSERT INTO RpAnimationController (
                RpAnimationControllerFile_fk, identifier, jsonPath
            ) VALUES (?, ?, ?)
            ''',
            (file_pk, identifier_data, ac_walker.path_str)
        )
        rpanim_pk = cursor.lastrowid

        states = ac_walker / "states" // str
        # LOAD PARTICLE EFFECTS
        for particle_effect_walker in states / "particle_effects" // int:
            short_name = particle_effect_walker / "effect"
            if not isinstance(short_name.data, str):
                continue
            cursor.execute(
                '''
                INSERT INTO RpAnimationControllerParticleEffect (
                    RpAnimationController_fk, shortName, jsonPath
                ) VALUES (?, ?, ?)
                ''',
                (rpanim_pk, short_name.data, particle_effect_walker.path_str)
            )
        # LOAD SOUND EFFECTS
        for sound_effect_walker in states / "sound_effects" // int:
            short_name = sound_effect_walker / "effect"
            if not isinstance(short_name.data, str):
                continue
            cursor.execute(
                '''
                INSERT INTO RpAnimationControllerSoundEffect (
                    RpAnimationController_fk, shortName, jsonPath
                ) VALUES (?, ?, ?)
                ''',
                (rpanim_pk, short_name.data, sound_effect_walker.path_str)
            )



from sqlite3 import Connection
from pathlib import Path
from .better_json_tools import load_jsonc
import json

CLIENT_ENTITY_BUILD_SCRIPT = '''
-- Resource pack entity file & content
CREATE TABLE ClientEntityFile (
    ClientEntityFile_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ResourcePack_fk INTEGER,

    path Path NOT NULL,
    FOREIGN KEY (ResourcePack_fk) REFERENCES ResourcePack (ResourcePack_pk)
        ON DELETE CASCADE
);
CREATE INDEX ClientEntityFile_ResourcePack_fk
ON ClientEntityFile (ResourcePack_fk);

CREATE TABLE ClientEntity (
    ClientEntity_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ClientEntityFile_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    FOREIGN KEY (ClientEntityFile_fk) REFERENCES ClientEntityFile (ClientEntityFile_pk)
        ON DELETE CASCADE
);
CREATE INDEX ClientEntity_ClientEntityFile_fk
ON ClientEntity (ClientEntityFile_fk);

CREATE TABLE ClientEntityRenderControllerField (
    ClientEntityRenderControllerField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ClientEntity_fk INTEGER NOT NULL,

    identifier TEXT NOT NULL,
    condition TEXT,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (ClientEntity_fk) REFERENCES ClientEntity (ClientEntity_pk)
        ON DELETE CASCADE
);
CREATE INDEX ClientEntityRenderControllerField_ClientEntity_fk
ON ClientEntityRenderControllerField (ClientEntity_fk);

CREATE TABLE ClientEntityGeometryField (
    ClientEntityGeometryField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ClientEntity_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,
    
    FOREIGN KEY (ClientEntity_fk) REFERENCES ClientEntity (ClientEntity_pk)
        ON DELETE CASCADE
);
CREATE INDEX ClientEntityGeometryField_ClientEntity_fk
ON ClientEntityGeometryField (ClientEntity_fk);

CREATE TABLE ClientEntityTextureField (
    ClientEntityTextureField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ClientEntity_fk INTEGER NOT NULL,


    shortName TEXT NOT NULL,
    -- identifier is the path without the extension
    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (ClientEntity_fk) REFERENCES ClientEntity (ClientEntity_pk)
        ON DELETE CASCADE
);
CREATE INDEX ClientEntityTextureField_ClientEntity_fk
ON ClientEntityTextureField (ClientEntity_fk);


CREATE TABLE ClientEntityMaterialField (
    ClientEntityMaterialField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ClientEntity_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (ClientEntity_fk) REFERENCES ClientEntity (ClientEntity_pk)
        ON DELETE CASCADE
);
CREATE INDEX ClientEntityMaterialField_ClientEntity_fk
ON ClientEntityMaterialField (ClientEntity_fk);

CREATE TABLE ClientEntityAnimationField (
    ClientEntityAnimationField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ClientEntity_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (ClientEntity_fk) REFERENCES ClientEntity (ClientEntity_pk)
        ON DELETE CASCADE
);
CREATE INDEX ClientEntityAnimationField_ClientEntity_fk
ON ClientEntityAnimationField (ClientEntity_fk);

CREATE TABLE ClientEntityAnimationControllerField (
    ClientEntityAnimationControllerField_pk INTEGER PRIMARY KEY AUTOINCREMENT,
    ClientEntity_fk INTEGER NOT NULL,

    shortName TEXT NOT NULL,
    identifier TEXT NOT NULL,
    jsonPath TEXT NOT NULL,

    FOREIGN KEY (ClientEntity_fk) REFERENCES ClientEntity (ClientEntity_pk)
        ON DELETE CASCADE
);
CREATE INDEX ClientEntityAnimationControllerField_ClientEntity_fk
ON ClientEntityAnimationControllerField (ClientEntity_fk);
'''

def load_client_entities(db: Connection, rp_id: int):
    rp_path: Path = db.execute(
        "SELECT path FROM ResourcePack WHERE ResourcePack_pk = ?",
        (rp_id,)
    ).fetchone()[0]

    for entity_path in (rp_path / "entity").rglob("*.json"):
        load_client_entity(db, entity_path, rp_id)

def load_client_entity(db: Connection, entity_path: Path, rp_id: int):
    cursor = db.cursor()
    # ENTITY FILE
    cursor.execute(
        "INSERT INTO ClientEntityFile (path, ResourcePack_fk) VALUES (?, ?)",
        (entity_path.as_posix(), rp_id))

    file_pk = cursor.lastrowid
    try:
        entity_jsonc = load_jsonc(entity_path)
    except json.JSONDecodeError:
        # sinlently skip invalid files. The file is in db but has no data
        return
    description = entity_jsonc / "minecraft:client_entity" / "description"

    # ENTITY - IDENTIFIER
    identifier = (description / "identifier").data
    if not isinstance(identifier, str):
        identifier = None
    cursor.execute(
        '''
        INSERT INTO ClientEntity (
        identifier, ClientEntityFile_fk
        ) VALUES (?, ?)
        ''',
        (identifier, file_pk))
    entity_pk = cursor.lastrowid
    # RENDER CONTROLLERS - unconditional
    for rc in (description / "render_controllers" // int):
        if isinstance(rc.data, str):
            identifier = rc.data
        else:
            continue  # Probably conditional render controller
        cursor.execute(
            '''
            INSERT INTO ClientEntityRenderControllerField (
                ClientEntity_fk, identifier, jsonPath
            ) VALUES (?, ?, ?)
            ''',
            (entity_pk, identifier, rc.path_str))
    # RENDER CONTROLLERS - conditional
    for rc in (description / "render_controllers" // int // str):
        if isinstance(rc.data, str):
           condition = rc.data
        else:
            condition = None
        cursor.execute(
            '''
            INSERT INTO ClientEntityRenderControllerField (
                ClientEntity_fk, identifier, condition, jsonPath
            ) VALUES (?, ?, ?, ?)
            ''',
            (entity_pk, rc.parent_key, condition, rc.path_str)
        )
    # MATERIALS
    for material in description / "materials" // str:
        if isinstance(material.data, str):
            identifier = material.data
        else:
            identifier = None
        cursor.execute(
            '''
            INSERT INTO ClientEntityMaterialField (
                ClientEntity_fk, shortName, identifier, jsonPath
            ) VALUES (?, ?, ?, ?)
            ''',
            (entity_pk, material.parent_key, identifier, material.path_str))
    # TEXTURES
    for texture in description / "textures" // str:
        if isinstance(texture.data, str):
            identifier = texture.data
        else:
            identifier = None
        cursor.execute(
            '''
            INSERT INTO ClientEntityTextureField (
                ClientEntity_fk, shortName, identifier, jsonPath
            ) VALUES (?, ?, ?, ?)
            ''',
            (entity_pk, texture.parent_key, identifier, texture.path_str))
    # GEOMETRIES
    for geometry in description / "geometry" // str:
        if isinstance(geometry.data, str):
            identifier = geometry.data
        else:
            identifier = None
        cursor.execute(
            '''
            INSERT INTO ClientEntityGeometryField (
                ClientEntity_fk, shortName, identifier, jsonPath
            ) VALUES (?, ?, ?, ?)
            ''',
            (entity_pk, geometry.parent_key, identifier, geometry.path_str))
    # ANIMATIONS & ANIMATION CONTROLLERS
    for animation in description / "animations" // str:
        if isinstance(animation.data, str):
            identifier = animation.data
        else:
            continue
        if animation.data.startswith("controller.animation."):
            # Animations
            cursor.execute(
                '''
                INSERT INTO ClientEntityAnimationControllerField (
                    ClientEntity_fk, shortName, identifier, jsonPath
                ) VALUES (?, ?, ?, ?)
                ''',
                (entity_pk, animation.parent_key, identifier, animation.path_str))
        elif animation.data.startswith("animation."):
            # Animation Controllers

            cursor.execute(
                '''
                INSERT INTO ClientEntityAnimationField (
                    ClientEntity_fk, shortName, identifier, jsonPath
                ) VALUES (?, ?, ?, ?)
                ''',
                (entity_pk, animation.parent_key, identifier, animation.path_str))
